/**
 * Author: Pat Beirne
 *
 * Test program for video capture and tv ouput.
 * Uses the Video4Linux2 driver
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <linux/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#include <assert.h>

#include <vga.h>
#include "videodev.h"
#include "vidcap.h"

int cmd;
void DoCapture(int handle, int palette);
void PrintProps(int);
void MovingWindow(int);
void FreezingWindow(int);
void SizingWindow(int);
void RunSettings(int);

int DoMenu(void);
void MovingWindow(int);
int	init_graph(void);
void	close_graph(void);
void shutdown(void);
int waitkey(void);
int     h_vidcap;
char *pSaveScreen;

#define MAX_CAP_WIDTH 704
#define MAX_CAP_HEIGHT 480

int main(int argc, char** argv)
{
    char c;
    int done = 0;

    /* change me to getopt() */
    if (argc>1) {
        c = argv[1][0];
        if (c=='-') c = argv[1][1];
        if (c=='-') c = argv[1][2];
        
        switch(c) {
        case 'h':
        case 'H':
            printf("usage: new_test [c] [y] [m]\n");
            printf("   x    operate within X windows\n");
            printf("   c    capture to file cnn.ppm\n");
            printf("   y    capture to file cnn.yuv\n");
            printf("   m    move capture window around screen\n");
            printf("        other, full test sequence:\n");
            printf("          blank, show, resize, move, blink, freeze\n");
            exit(0);
            break;
        default:
            cmd = c;
            break;
        }
    }

    h_vidcap = open_vidcap();

    atexit(shutdown);

    set_text_mode();

    PrintProps(h_vidcap);
    while(c = DoMenu()) {
        /* eat the enter key */
        vga_getkey();
        
        if (c=='q' || c=='Q')
            exit(0);

        switch(c) {
        case 'c':
            DoCapture(h_vidcap, PIX_FMT_RGB24);
            break;
        case 'y':
            DoCapture(h_vidcap, PIX_FMT_YUYV);
            break;
        case '5':
            DoCapture(h_vidcap, PIX_FMT_RGB565);
            break;
        case 'm':
            MovingWindow(h_vidcap);
            break;
        case 's':
            RunSettings(h_vidcap);
            break;
        case 'z':
            SizingWindow(h_vidcap);
            break;
        case 'f':
            FreezingWindow(h_vidcap);
            break;
        default:
            exit(1);
        }
        set_text_mode();
    }

    
}

/* try to open the video device,
 * if that fails, it may be because we are not in graphics mode
 * try to open the svgalib graphics mode
 * if that works, establish an area for video capture and
 *   save away any bytes there
 * several of these cases will fail in a fatal way
 */
int open_vidcap(void)
{
    int h_vidcap = open("/dev/video0",O_RDWR);
    if (h_vidcap<0) {
        printf("errno %d\n",errno);
        /* try again in graphics mode */
        if (errno == ENXIO) {
            if (init_graph() < 0)
                /* init_graph() reports its own errors */
                exit(1);
            /* save away the extra memory */
            {
                /* make an area in video memory for our capture;
                 instead of a fixed 1Meg offset, we really should calculate
                 an offset and use that */
                struct video_framebuffer vb = { 0, 0, {(void*)0xd0100000},
                { MAX_CAP_HEIGHT, MAX_CAP_WIDTH, 2, PIX_FMT_YUYV} };
                int hMem;

                /* try again */
                h_vidcap = open("/dev/video0",O_RDWR);
                if (h_vidcap<0) {
                    printf("error: cannot open /dev/video0 from graphics mode\n");
                    exit(1);
                }
                /* ok, we have svglib graphics; save away the capture ram */
                pSaveScreen = (char*)malloc(MAX_CAP_WIDTH * MAX_CAP_HEIGHT * 2);
                hMem = open("/dev/mem",O_RDONLY);
                printf("hMem %d\n",hMem);
                printf("seek %p\n",lseek(hMem, 0xd0100000, SEEK_SET));
                printf("read %d\n",read(hMem, pSaveScreen, MAX_CAP_WIDTH * MAX_CAP_HEIGHT * 2));
                close(hMem);
                ioctl(h_vidcap, VIDIOC_S_FBUF, &vb);
            }
        } else {
            printf("error: cannot open /dev/video0\n");
            exit(1);
        }
    }
    return h_vidcap;
}


void shutdown(void) {
    if (h_vidcap>=0)
        close(h_vidcap);
    close_graph();
    printf("shutdown done\n");
}

void
DoCapture(int h, int palette)
{
    FILE* f_write;
    struct video_format param;
    struct video_fmtdesc vf;
    int r;
    char* p;
    int multiply = (palette == PIX_FMT_RGB24 ? 3 : 2);

    printf("DoCapture\n");
    printf("The system will now capture an image in %s format\n",
           multiply==3 ? "RGB" :
           (palette == PIX_FMT_YUYV ? "YUV" : "RGB16"));

    waitkey();

    set_graph_mode();
    param.width = 352;
    param.height = 240;
    param.pixelformat = palette;
    printf("palette of %x\n",palette);
    param.depth = multiply*8;
    r = ioctl(h, VIDIOC_S_FMT, &param);
    if (r) {
        printf("error: capture attempt returned %d\n",r);
        exit(1);
    }
    assert(param.width * param.height * multiply == param.sizeimage);
    if (multiply==3)
        f_write = fopen("cnn.ppm","w");
    else if (palette == PIX_FMT_YUYV)
        f_write = fopen("cnn.yuv","w");
    else
        f_write = fopen("cnn.rgb16","w");
    fprintf(f_write,"P6\n%d %d\n255\n",param.width,param.height);
    p = (char*) malloc(param.width * param.height * multiply);
    sleep(1);
    read(h,p,param.width * param.height * multiply);
    fwrite(p,param.width * param.height * multiply, 1, f_write);
    fclose(f_write);
    
    printf("capture file written to cnn.ppm\n");
}

void PrintProps(int h)
{
    struct video_capability vc;
    struct video_fmtdesc vf;
    struct video_queryctrl vq;
    int r;
    
    ioctl(h, VIDIOC_G_CAP, &vc);
    printf("The %s capture device can:\n",vc.name);
    if (vc.type & VID_TYPE_CAPTURE) printf("\tcapture video images\n");
    if (vc.type & VID_TYPE_TUNER)   printf("\ttune in tv stations\n");
    if (vc.type & VID_TYPE_TELETEXT)printf("\tcapture teletext data\n");
    if (vc.type & VID_TYPE_PREVIEW) printf("\tdisplay the capture on screen\n");
    printf("It has %d inputs for video and %d inputs for audio\n",vc.inputs,vc.audios);
    printf("The largest frame that it can capture is %d by %d\n\n",
           vc.maxwidth,vc.maxheight);

    printf("This capture system is capable of the following formats:\n"
           "Name                           Format Depth\n");
    r = 0;
    for (vf.index=0; r==0; vf.index++) {
        r = ioctl(h, VIDIOC_G_CAPFMT, &vf);
        if (r==0) {
            printf("%-31s%-8s%d\n",vf.description,&vf.pixelformat,vf.depth);
        }
    }

    printf("The display part of the capture system is capable of the following formats:\n"
           "Name                           Format Depth\n");
    r = 0;
    for (vf.index=0; r==0; vf.index++) {
        r = ioctl(h, VIDIOC_G_FBUFFMT, &vf);
        if (r==0) {
            printf("%-31s%-8s%d\n",vf.description,&vf.pixelformat,vf.depth);
        }
    }
    printf("\n");

    printf("The driver has ajustments for:\n");
    printf("  Label        type    Minimum  Maximum    Step  Default\n");
     
    r = 0;
    for (vq.id=CID_BRIGHTNESS; vq.id<CID_BRIGHTNESS+4; vq.id++) {
        r = ioctl(h, VIDIOC_QUERYCTRL, &vq);
        if (r==0)
            printf("  %-12s %-6s %8d %8d %7d %8d\n",
	       vq.name,
	       (vq.step == -1) ? "(enum)" : (vq.step == 0) ? "(bool)" :
	       "(int)",
                   vq.minimum, vq.maximum, vq.step, vq.default_value);
    }

    printf("\n");
}

int DoMenu(void) {
    printf("Please enter a key to test a feature:\n");
    printf("c    capture RGB to the file cnn.ppm\n");
    printf("y    capture YUV to the file cnn.yuv\n");
    printf("5    capture RGB16 to the file cnn.rgb16\n");
    printf("m    move a window around the screen, with hide & show\n");
    printf("z    resize a window\n");
    printf("s    run through brightness, contrast, hue & saturation sweeps\n");
    printf("f    freeze and unfreeze capture\n");
    printf("While the test is run, you may need to hit <Enter> to proceed.\n");
    return waitkey();
}
void MovingWindow(int h)
{
    int i,j;
    struct video_window vw = {0,0,352,240,0,0,0};

    printf("This test will move the display version of the video capture\n"
       "around the screen, first smoothly, then to random locations.\n");
    waitkey();
    
    set_graph_mode();
    ioctl(h, VIDIOC_S_WIN, &vw);
    ioctl(h, VIDIOC_PREVIEW, 1);
    system("beep");
    for (i=0; i<640-352; i++) {
        vw.x = i;
        ioctl(h, VIDIOC_S_WIN, &vw);
        usleep(10000);
    }
    waitkey();
    for (i=640-352; i>0; i--) {
        vw.x = i;
        vw.y++;
        ioctl(h, VIDIOC_S_WIN, &vw);
        usleep(10000);
    }
    waitkey();
    for (i=0; i<25; i++) {
        vw.x = (640-352)*(rand()>>8)/(RAND_MAX>>8);
        vw.y = (480-240)*(rand()>>8)/(RAND_MAX>>8);
        ioctl(h, VIDIOC_S_WIN, &vw);
        usleep(50000);
    }
    waitkey();
    ioctl(h, VIDIOC_PREVIEW, 0);
}

void
FreezingWindow(int h) {
    int i,j;
    struct video_window vw = {0,0,352,240,0,0,0};
    printf("This test shows a window, and freezes it, enlarges it and unfreezes\n");
    waitkey();
    set_graph_mode();
    
    ioctl(h, VIDIOC_S_WIN, &vw);
    ioctl(h, VIDIOC_PREVIEW, 1);
    waitkey();
    waitkey();
    ioctl(h, CMD_FREEZE);
    waitkey();
    vw.width = 400;
    vw.height = 300;
    ioctl(h, VIDIOC_S_WIN, &vw);
    waitkey();
    ioctl(h, CMD_UNFREEZE);
    waitkey();
    ioctl(h, VIDIOC_PREVIEW, 0);

}

void
SizingWindow(int h) {
    int i,j;
    struct video_window vw = {0,0,352,240,0,0,0};

    printf("This test runs the display window from very small to very large\n");
    printf("Note that the capture size is 352x240, and that the display will not\n");
    printf("go smaller than the capture, but can go bigger\n");
    set_graph_mode();

    ioctl(h, VIDIOC_S_WIN, &vw);
    ioctl(h, VIDIOC_PREVIEW, 1);
    waitkey();
    for (i=32; i<640; i+=32) {
        vw.width = i;
        vw.height = 3 * i / 4;
        ioctl(h, VIDIOC_S_WIN, &vw);
        usleep(900000);
    }
    waitkey();
}

void
RunSettings(int h) {
    struct video_queryctrl vq;
    struct video_control vc;
    int i,r, orig;
    struct video_window vw = {0,0,352,240,0,0,0};

    printf("This test will run through brightness, contrast, saturation and hue ranges\n");
    printf("and restore the original values after the sweep\n");
    waitkey();
    /* display a plain window */
    set_graph_mode();
    ioctl(h, VIDIOC_S_WIN, &vw);
    ioctl(h, VIDIOC_PREVIEW, 1);

    vc.id = vq.id = CID_BRIGHTNESS;
    r = ioctl(h, VIDIOC_QUERYCTRL, &vq);
    if (r==0) {
        ioctl(h, VIDIOC_G_CTRL,&vc);
        orig = vc.value;
        for (i=vq.minimum; i<vq.maximum; i+=vq.step) {
            vc.id = vq.id;
            vc.value = i;
            ioctl(h,VIDIOC_S_CTRL,&vc);
        }
        vc.value = orig;
        ioctl(h,VIDIOC_S_CTRL,&vc);
    }
    waitkey();
    vc.id = vq.id = CID_CONTRAST;
    r = ioctl(h, VIDIOC_QUERYCTRL, &vq);
    if (r==0) {
        ioctl(h, VIDIOC_G_CTRL,&vc);
        orig = vc.value;
        for (i=vq.minimum; i<vq.maximum; i+=vq.step) {
            vc.id = vq.id;
            vc.value = i;
            ioctl(h,VIDIOC_S_CTRL,&vc);
        }
        vc.value = orig;
        ioctl(h,VIDIOC_S_CTRL,&vc);
    }
    waitkey();
    vc.id = vq.id = CID_SATURATION;
    r = ioctl(h, VIDIOC_QUERYCTRL, &vq);
    if (r==0) {
        ioctl(h, VIDIOC_G_CTRL,&vc);
        orig = vc.value;
        for (i=vq.minimum; i<vq.maximum; i+=vq.step) {
            vc.id = vq.id;
            vc.value = i;
            ioctl(h,VIDIOC_S_CTRL,&vc);
        }
        vc.value = orig;
        ioctl(h,VIDIOC_S_CTRL,&vc);
    }
    waitkey();
    vc.id = vq.id = CID_HUE;
    r = ioctl(h, VIDIOC_QUERYCTRL, &vq);
    if (r==0) {
        ioctl(h, VIDIOC_G_CTRL,&vc);
        orig = vc.value;
        for (i=vq.minimum; i<vq.maximum; i+=vq.step) {
            vc.id = vq.id;
            vc.value = i;
            ioctl(h,VIDIOC_S_CTRL,&vc);
        }
        vc.value = orig;
        ioctl(h,VIDIOC_S_CTRL,&vc);
    }
    waitkey();
    ioctl(h, VIDIOC_PREVIEW, 0);
}

int waitkey(void) {
    int c;
    while (!(c=vga_getkey()));
    return c;
}
    
#if 0

static const char*	hello = "Hello how are you?";
char			greetings[20];
//unsigned char		RGBBuf[352*240*3];
unsigned char		RGBBuf[704*480*3];
unsigned char		YUVBuf[352*240*2];
unsigned char*		lpVGAMem = 0;
struct vidcap_param	test;
int			fd = 0;
int			bmp_fd = 0;

#define XLATTABSIZE	256     /*xlat table size*/
short  			XlatY[XLATTABSIZE] 	= { 0 };
short  			XlatU_B[XLATTABSIZE] 	= { 0 };
short  			XlatU_G[XLATTABSIZE] 	= { 0 };
short  			XlatV_G[XLATTABSIZE] 	= { 0 };
short  			XlatV_R[XLATTABSIZE] 	= { 0 };

#define min(a,b)			(a<b?a:b)
#define max(a,b)			(a>b?a:b)
#define CVHI					255
#define CVLO					0

#include "yuvtorgb.c"

time_t	time0;
int	frame = 0;

int main(int argc, void* argv)
{

    int i;
            /* create a parameter to pass to the ioctls */
    struct vidcap_param	rect = { 10, 10, 352, 288, 0, 0 };


        /* an argv will skip the vgalib init: capture in text mode */
        if (argc<2)
	{
	}

        /* open the device */
        if (open_vidcap() < 0)
		return;

	//vga_screenoff();
	

        /* decorate a background */
        if (argc<2)
        {
            if (open_bmp() < 0)
		return;
            show_bmp(bmp_fd);
            close_bmp();
            /* make a background for our capture */
            gl_fillbox(rect.x, rect.y, rect.width, rect.height, 0);
        }
        vga_getkey();
        
        /* for local YUV->RGB translations, init the table */
	//IGS_InitXlatTab();

        /* set the chroma key color to black */
        rect.lpRGBBuf = 0;
        /* start a simple video capture */
        ioctl(fd, CMD_CAPTURE, &rect);

        // to demonstrate various sizes of capture
        if (0)
        {
            int i;
            vga_getkey();
            gl_fillbox(0,0,640,480,0);
            for (i=32; i<640; i+=32) {
                rect.width = i;
                rect.height = 3 * i / 4;
                ioctl(fd, CMD_CAPTURE, &rect);
                sleep(1);
            }
        }
	
        //vga_screenon();
        
        /* freeze the image.....optional */
//	ioctl(fd, CMD_FREEZE);
	
        vga_getkey();

        /* a loop to demonstrate brightness & contrast */
        if (0)
        {
            struct vidcap_pict;
            for (i=0; i<255; i++) {
                ioctl(fd, VIDIOCGPICT, &pict);
                pict.brightness += 0x100;
                ioctl(fd, VIDIOCSPICT, &pict);
                usleep(1000);
            }
        }

        /* the capture can be done invisibly
         * this also waits for V-sync, so the capture is complete
         */
//	ioctl(fd, CMD_HIDE);
//	getchar();
//	ioctl(fd, CMD_SHOW);

//	getchar();
//	ioctl(fd, CMD_HIDE);


	/* To enable the second overlay using the Graphics Engine Registers.
	 * IMPORTANT:
	 *			Please check the values for Graphics Engine Register Base
	 *			Address in igscop.h. Search for XGAAddr #define.
	 */
//	ioctl(fd, CMD_CAPTURE2);

	//getchar();
	//ioctl(fd, CMD_RELEASE2);

//	yuv2rgb_dump();

//	yuv_dump(&test);

//	capture_rate();

        /* transfer the image from vidcap device to local ram */
        yuv_capture(&rect);

        //yuv_capture_and_dump2(&rect);

	//vga_setrgbcolor(255, 0, 0);
	//vga_drawline(300, 250, 340, 250);
	//getchar();
	//ioctl(fd, CMD_RELEASE2);
	//ioctl(fd, CMD_RELEASE);

        getchar();
        ioctl(fd, CMD_FREEZE);
        rect.x = 30;
        rect.y = 100;
        rect.width = 25;
        rect.height = 45;
	ioctl(fd, CMD_START_PIP, &rect);		/* Start PIP */
	ioctl(fd, CMD_UNFREEZE);
	getchar();

	//ioctl(fd, CMD_CAPTURE2);

	//ioctl(fd, CMD_RELEASE2);
	ioctl(fd, CMD_END_PIP);
	ioctl(fd, CMD_RELEASE);

	close_vidcap();
        if (argc<2)
	{
            close_graph();
	}
	
	printf("frames = %d\n", frame);
}
#endif

int init_graph()
{
	if (vga_init() != 0)
	{
		printf("vga_init failed\n");
		return -1;
	}

	if (vga_hasmode(G640x480x16M) != 1)
	{
		printf("vga_hasmode failed\n");
		return -1;
	}
	
	if (vga_setmode(G640x480x16M) != 0)
	{
		printf("vga_setmode failed\n");
		return -1;
	}
	return 0;
}

int set_graph_mode()
{
    if (pSaveScreen && vga_getcurrentmode()==TEXT)
        vga_setmode(G640x480x16M);
}
int set_text_mode()
{
    if (pSaveScreen && vga_getcurrentmode()!=TEXT)
        vga_setmode(TEXT);
}
void close_graph()
{
    if (pSaveScreen) {
        /* saved memory must be restored in graphics mode */
        int hMem = open("/dev/mem",O_WRONLY);
        set_graph_mode();
        lseek(hMem, 0xd0100000, SEEK_SET);
        write(hMem, pSaveScreen, MAX_CAP_WIDTH * MAX_CAP_HEIGHT * 2);
        close(hMem);
        set_text_mode();
    }
}

#if 0
int open_vidcap()
{
	fd = open("/dev/vidcap", O_RDONLY);
	if (fd < 0)
		printf("Error opening /dev/vidcap\n");

	return fd;
}

void close_vidcap()
{
	close(fd);
}

int open_bmp()
{
	bmp_fd = open("video640.bmp", O_RDONLY);
	if (bmp_fd < 0)
		printf("Error opening Bitmap File.\n");

	return bmp_fd;
}

void close_bmp()
{
	close(bmp_fd);
}

/* To dump YUV data frame at off-screen memory address
 * to confirm validity of video capture.
 OLD CODE! PROBABLY WILL NOT WORK patb 25/09/98
 */
void yuv_dump(struct vidcap_param* test)
{
	//FILE*	dump_file = fopen("/mnt/mrc.milindc/oneframe.dat", "r");
	FILE*	dump_file = fopen("cnn.dat", "r");
	if (dump_file)
	{
		int	row = 0;
		test->lpRGBBuf = YUVBuf;
		while (!feof(dump_file))
		{
			fread(YUVBuf, 352*2*240, 1, dump_file);
		#if 1
			ioctl(fd, CMD_DUMP, test);
		#else
			ConvertYUVtoRGB24(RGBBuf, YUVBuf, 352, 240);
			for (row = 0; row != 240; row++)
				vga_drawscansegment(RGBBuf + (row*352*3), 0, row, 352*3);
		#endif
				
			//getchar();
		}
		fclose(dump_file);
	}
	else
		//printf("vidcap_test: Error opening /mnt/mrc.milindc/cnn.dat\n");
		printf("vidcap_test.c: Error opening oneframe.dat\n");
}
#endif

#if 0
/* OLD CODE! PROBABLY WILL NOT WORK patb 25/09/98 */
void yuv2rgb_dump()
{
	/* To capture frames and receiving YUV -> RGB converted
	 * data.
	 * Dumping the frame on the vga on-screen memory.
	 */
	while (!vga_getkey()) {
		if (ioctl(fd, CMD_GET_DATA, &test) == 0)
		{
			int	row = 0;

			for (row = 0; row != 240; row++)
				vga_drawscansegment(test.lpRGBBuf + ((row)*352*3), 0, row, 352*3);
		}
		else
		{
			printf("vidcap_test: Error in CMD_GET_DATA.\n");
			break;
		}
	}// while(!vga_getkey)

	ioctl(fd, CMD_RELEASE);
}

void capture_rate()
{
	/* To calculate the max rate for reading frames. */
	//getchar();
	FILE*	dump_file = fopen("cnn.dat", "w");
	time0 = time(0);
	while(difftime(time(0), time0) < 2)
	{
		ioctl(fd, CMD_GET_DATA, &test);
		fwrite(test.lpYUVBuf, 352*2*240, 1, dump_file);
		frame++;
	}
	fclose(dump_file);
}

void yuv_capture(struct vidcap_param * param)
{
	/* To capture frames and dump it to a file. */
        FILE*	dump_file = fopen("cnn.dat", "w");
        param->lpYUVBuf = 0;
       	param->lpRGBBuf = RGBBuf;

        if (dump_file)
	{
	    int	total = 50;
	    time0 = time(0);
            ioctl(fd, CMD_GET_DATA, param);
            fprintf(dump_file,"P6\n%d %d\n%d\n",param->width,param->height,255);
            fwrite(param->lpRGBBuf, param->width*param->height*3, 1, dump_file);
            frame++;
            fclose(dump_file);
       	} else
            printf("vidcap_test.c: Error opening cnn.dat\n");
}

/* OLD CODE! PROBABLY WILL NOT WORK patb 25/09/98 */
void yuv_capture_and_dump2(struct vidcap_param* test)
{
	int	iterations = ITERATIONS;

	time0 = time(0);
	while(difftime(time(0), time0) < 2)
	//while (iterations--)
	{
		ioctl(fd, CMD_GET_DATA, test);
		ioctl(fd, CMD_DUMP, test);
		frame++;
	}
}
#endif

//#include "gramem.c"

