#ifndef _VIDCAP_H
#define _VIDCAP_H

#include "videodev.h"

/************************************
 the IOCTL commands; see the man page
 ************************************/

/* most of the ioctls are in videodev.h */

#define CMD_FREEZE             _IOW('v', BASE_VIDIOC_PRIVATE+3, void)
#define CMD_UNFREEZE           _IOW('v', BASE_VIDIOC_PRIVATE+4, void)

#define VERSION 0x00000201 /* in english: 2.1 */


/* a typical sequence is:
 open()
 CMD_CAPTURE
 CMD_FREEZE
 read()
 CMD_RELEASE
 close()

 open()
 Checks for graphic mode
 
 CMD_CAPTURE
 Starts the capture action. Assumes that it owns the capture memory.

 It is legal to call CMD_CAPTURE repeatedly without a corresponding CMD_RELEASE
 as long as the CMD_RELEASE is called eventually

 The parameter describes a rectangle for capture. This also defines
 the video memory size required to hold the capture. IT IS THE RESPONSIBILITY
 OF THE CALLER to keep the memory size below the system requirements. For
 instance, if the user has 2Megs of memory, and is running 800x600x3, then
 that leaves 657k. You should not capture any more than this, which is
 about 640x480.

 This idiom does not display anything; only capture.

 Another typical sequence is:
 open()
 VIDIOCGCAP
 VIDIOCSWIN
 VIDIOCSPICT
 VIDIOCCAPTURE
 read()
 close()

 This sequence only uses the public device driver calls (VIDIOC*)
 This sequence captures and displays on screen.
 This sequence makes assumptions about the capture rectangle.

 open()
 Check the graphic mode

 VIDIOCGCAP
 [optional] checks capabilities; yes we can capture & overlay

 VIDIOCSWIN
 set a graphics window for capture; since CMD_CAPTURE was not issued,
 the capture rectangle is set to the same size as the overlay (displayed)
 window

 VIDIOCSPICT
 set the brightness, contrast & capture_data format

 VIDIOCAPTURE
 enable both capture and overlay

 read()
 gets the bytes from the capture buffer; the size is set by the
 most recent CMD_CAPTURE (or, if there has been no CMD_CAPTURE, by
 the most recent VIDIOSWIN)

 close()
 releases the resources; turns off overlay and capture

 NOTE: the CMD_CAPTURE no longer sets the background for the window; it
 is up to the caller to set the screen color.

 Private IOCTLS
 
 CMD_RELEASE
 Turns off the capture

 CMD_HIDE
 Inhibits the display of the capture data on the screen.

 CMD_SHOW
 Displays the capture data on the screen.

 The CMD_SHOW & CMD_HIDE both invoke a one-vga-frame delay; the bad news is
 that it slows down your Linux box; the good news is that a full camera frame is
 probably sampled. Tearing on the VGA is avoided.
 */

/************************************
 the structures; see the man page
 ************************************/


/* this struct is used by CMD_GET/SET_SCREEN_PARAMS */
struct vidcap_screen_params {
        int x_ext, y_ext;       /* horiz/vert total pixels */
        int x_orig, y_orig;     /* offset to displayed pixels */
        int width, height;      /* size of displayed pixels */
        int bytes_per_pixel;    /* 1, 2 or 3 */
        int total_video_memory; /* 1, 2 or 4Megs */
        int display_memory_used; /* memory used by x_ext * y_ext * bpp */
        int offset_vidcap_mem;
        int offset_overlay2_mem;
};


/**************** PRIVATE ****************/
struct vidcap_size {
    int         width, height;
};
struct vidcap_point {
    int         x,y;
};
struct vidcap_rect {
    int         x,y;
    int         width,height;
};


/* the device structure; holds current state information */
struct nw_cap_device
{
    /* module managment */
    int                    open_count;
    int                    minor;

    /* capture state;
     only the realized info is kept; the requested info is discarded */
    struct video_format    capture_format;
    struct video_parm      capture_param;      /* contains xflip, yflip, hires */
    int                    control[4];         /* brightness, contrast, saturation & hue */
    struct vidcap_size     camera_data_size;   /* the bits coming from the camera */
    int                    capturing:1;
    int                    capture_format_set:1;   /* true if cap_params were set */
    int                    videoStandard:8;
    
    /* display state */
    struct vidcap_rect     requested_window;
    struct vidcap_rect     realized_window;
    int                    chromakey;
    struct vidcap_screen_params screen;         /* current vga settings */
    int                    uses_chromakey:1;
    int                    displaying:1;
    int                    xMirrorDisplay:1;
    int                    yMirrorDisplay:1;

    /* mmap managment */
    int                    mem_buf_set:1;  /* if the user set the memory address */
    int                    stream_buffers_mapped:1; /* if the buffers have been memmapped */

    /* performance stuff */
    struct video_performance perform;
};
/************************************
 the constants; see the man page
 ************************************/

/* the defaults are the 0 cases */

/* these are extra flags for the struct video_window.flags word */
#define VIDEO_WINDOW_CHROMAKEY  0
#define VIDEO_WINDOW_WINDOWKEY  0x100

#define VIDCAP_MINOR_CAPTURE 0
#define VIDCAP_MINOR_OVERLAY 4
#define VIDCAP_MINOR_BACKDOOR 128

#define CONTROL_BRIGHTNESS 0
#define CONTROL_CONTRAST 1
#define CONTROL_SATURATION 2
#define CONTROL_HUE 3

#undef outb_p
#define outb_p(v, p) (*(volatile char*)(pIO + p) = v)
#undef inb_p
#define inb_p(p) (*(volatile unsigned char*)(pIO + p))
#undef outw_p
#define outw_p(v, p) outb_p(((char)(v&0xff)), (p)); \
                     outb_p(((char)((v&0xff00)>>8)), (p+1));

#endif /* _VIDCAP_H */
